"""
sharedstate

Shared State module.  Get access to shared state.

All state has a namespace, and then the state id.  sharecounter is used to
procure new state ids, so it can be incremented atomically and shared with
other procblock programs.

TODO(g): Implement serialization, archiving, snapshotting, and replication to
sharestate.  This will allow us flexibility in many things.  This is not
necessarily going to be the best scaling solution, but it will work and provide
a way to keep state and distribute.  Name spaces should have this specified
individually.  If not specified, state will not be archived, serialized,
snapshotted or replicated.

  * Use the "archive" and "snapshot" modules for this, so archival and
      snapshotting is universal.  Ensure a process restart will do the right
      thing in attempting to restore from snapshot, then archive, if present.
      
      TODO(g): Merge achive and snapshot.  They are the same technology.  If we
          really want to keep state, then we must archive each transactions and
          snapshot to avoid having to replay too many archives.
"""


import threading
import time
import run
import yaml
import os
import logging
import glob
import re

from log import log

import threadsafedict
import threadsafelist


# Limit the maximum depth search through the shared state will go
MAX_SEARCH_DEPTH = 10

# Global for state buckets.  Each bucket is dict and contains keys in a dict.
#   Temporary?
#TODO(g): Convert to threadsafe and use a better store than module, so it
#   survives reloads.
#TODO(g): Why not allow state to be passed in and store the locks there?
#   So state management and serialization can be dealt with someone else.
SHARED_STATE = threadsafedict.ThreadSafeDict()

# When setting or getting a value in a bucket, use locking to ensure this is
#   an atomic operation
SHARED_STATE_LOCKS = threadsafedict.ThreadSafeDict_IgnoreOverwrites()


class StateBucketDoesntExist(Exception):
  """This state bucket does not exist, so it cannot by returned."""


class StateBucketNotFound(Exception):
  """This state bucket was not found.  Different than it having a value of None."""
  
  
class StateKeyNotFound(Exception):
  """This state key was not found.  Different than it having a value of None."""


def _LockState(bucket):
  global SHARED_STATE_LOCKS
  
  if bucket not in SHARED_STATE_LOCKS:
    #NOTE(g): This is a race condition, which is why we use the
    #   ThreadSafeDict_IgnoreOverwrites class, which will ignore multiple
    #   attempts to create locks.  Only the first lock is created, and then
    #   will be safely used.  Any additional attempts to create locks will be
    #   ignored.
    SHARED_STATE_LOCKS[bucket] = threading.Lock()
  
  # Lock it
  SHARED_STATE_LOCKS[bucket].acquire()


def _UnlockState(bucket):
  global SHARED_STATE_LOCKS
  
  # Throw an exception if this counter doesnt exist, something is wrong
  if bucket not in SHARED_STATE_LOCKS:
    raise StateDoesntExist('Not found: %s' % bucket)
  
  # Release it
  SHARED_STATE_LOCKS[bucket].release()


def _EnsureBucketExists(bucket):
  global SHARED_STATE
  
  # Ensure no one else can operate on this bucket
  _LockState(bucket)
  
  # Test that the bucket does not exist
  if bucket not in SHARED_STATE:
    # Create the bucket
    SHARED_STATE[bucket] = threadsafedict.ThreadSafeDict()
  
  # Release the bucket, others may attempt this call now, or others
  _UnlockState(bucket)


def _Search_Dict(data, value, depth=0):
  """Search for the value in data (dict)."""
  #log('Search Dict: %s: %s' % (depth, str(data)[:25]))
  results = []
  
  # If we hit max search depth, return empty
  if depth > MAX_SEARCH_DEPTH:
    return []
  
  # We have not yet matched this data
  matched_data = False
  
  # Loop through the keys
  for key in data:
    key_data = data[key]
    
    #log('Key Data: %s: %s: %s' % (depth, key, str(key_data)[:25]))
    
    # If the key data is a dictionary, recurse and keep searching for
    #   matches!
    if type(key_data) == type({}) or isinstance(key_data, threadsafedict.ThreadSafeDict):
      results += _Search_Dict(key_data, value, depth=depth+1)
    
    # Else, this is a leaf of a dictionary, so add the dictionary itself and
    #   move on
    elif not matched_data and _Search_NonDict_IsMatch(key_data, value):
      matched_data = True
      results.append(data)
  
  return results


def _Search_NonDict_IsMatch(data, value):
  """Looks at data in a non-dictionary, and returns Boolean if anything mathches"""
  if type(data) == type([]) or isinstance(data, threadsafelist.ThreadSafeList):
  #if type(data) in ([], threadsafelist.ThreadSafeList):
    for item in data:
      if re.findall(value, str(item), re.IGNORECASE):
        #log('NonDict Match: %s: %s' % (value, str(data)[:25]))
        return True
    
    #log('NonDict NO Match: %s: %s' % (value, str(data)[:25]))
    return False
  
  # Else, this is not a list, so convert to string and test it
  else:
    if re.findall(value, str(data), re.IGNORECASE):
      #log('NonDict Match: %s: %s' % (value, str(data)[:25]))
      return True
    else:
      #log('NonDict NO Match: %s: %s' % (value, str(data)[:25]))
      return False


def RegisterDefaultSave(bucket, save):
  """Registers a default save file for a specified bucket."""
  # Register this bucket's save file
  Set('__sharedstate.save.registered', bucket, save)
  log('Registered Shared State Save: Bucket: %s  Save: %s' % (bucket, save))


def GetRegisteredSave(bucket):
  """If registered, returns the save file.  None if not found."""
  _EnsureBucketExists('__sharedstate.save.registered')
  
  if bucket in SHARED_STATE['__sharedstate.save.registered']:
    return SHARED_STATE['__sharedstate.save.registered'][bucket]
  else:
    return None


def Search(value, bucket_regex=None, key_regex=None):
  """Searchs through the specified bucket and key regexs."""
  results = []
    
  # Get all the shared state buckets
  buckets = GetBuckets()
  for bucket in buckets:
    if bucket_regex:
      match = re.findall(bucket_regex, bucket, re.IGNORECASE)
    else:
      match = True
    
    if match:
      # Get this buckets data
      bucket_data = GetBucketData(bucket)
      
      #log('Bucket: %s: %s: %s' % (bucket, type(bucket_data), str(bucket_data)[:25]))
      
      # If this bucket contains a dictionary
      if type(bucket_data) == type({}) or isinstance(bucket_data, threadsafedict.ThreadSafeDict):
        # Search the dictionary, and append the results (depth starts here)
        results += _Search_Dict(bucket_data, value, depth=0)
      
      #TODO(g): This would be cool, but I want to only display dicts in the
      #   Search dialog Data Table thing Im building at the moment, so holding
      #   off on this cool feature, so the other feature is easier to write.
      #elif _Search_NonDict_IsMatch(bucket_data, value):
      #  results.append(bucket_data)

  return results


def GetBuckets():
  """Returns a list of strings, names of buckets."""
  global SHARED_STATE
  
  return SHARED_STATE.keys()


def BucketExists(bucket):
  """A bucket exists.
  
  Args:
    bucket: string, name of bucket for state variables
  """
  global SHARED_STATE
  
  try:
    _LockState(bucket)
    
    exists = bucket in SHARED_STATE
    
    return exists

  finally:
    _UnlockState(bucket)
  

def KeyExists(bucket, key, strict=False):
  """A key exists inside a bucket.
  
  Args:
    bucket: string, name of bucket for state variables
    key: string or any, key to store data against.  Any valid dict key.
    string: boolean (default=False), if True, will raise StateBucketDoesntExist,
        if False, returns False, in case the bucket doesnt exist
  """
  global SHARED_STATE
  
  if not BucketExists(bucket):
    if strict:
      raise StateBucketDoesntExist('Bucket doesnt exist: %s' % bucket)
    else:
      return False
  
  exists = key in SHARED_STATE[bucket]
  
  return exists


def RemoveKey(bucket, key, strict=False, save=None, backups=2, delay=0,
              remote_node_sync=None, remote_node_url=None,
              remote_node_delay=None):
  """Removes a key from a bucket.
  
  Args:
    bucket: string, name of bucket for state variables
    key: string or any, key to store data against.  Will remove this key.
    string: boolean (default=False), if True, will raise StateBucketDoesntExist,
        if False, returns False, in case the bucket doesnt exist
  """
  global SHARED_STATE
  
  # If no save was specified, get the Registered save
  if save == None:
    save = GetRegisteredSave(bucket)
  
  if not BucketExists(bucket):
    if strict:
      raise StateBucketDoesntExist('Bucket doesnt exist: %s' % bucket)
    else:
      return False
  
  # If the key exists, remove it
  if key in SHARED_STATE[bucket]:
    # Remove the key from the shared state
    del SHARED_STATE[bucket][key]
    
    # If we want to save this key deletion
    if save:
      # %s == %(key)s
      if '%s' in save:
        save = save.replace('%s', '%(key)s')
      
      # If this key is saved into it's own file, dele it 
      if '%(key)s' in save:
        save_path = save % {'key':run.MakePathSafeToCreate(key)}
        os.unlink(save_path)
        log('Removed key save file: %s' % save_path)
      
      # Else, this key is part of a bucket file, so save that normally
      else:
        Save(bucket, key, save=save, backups=backups, delay=delay,
             remote_node_sync=remote_node_sync, remote_node_url=remote_node_url,
             remote_node_delay=remote_node_delay)
  
  # If the key doesnt exist, and were in strict mode, raise an exception
  #NOTE(g): strict==False means this returns True.
  elif strict:
    raise StateKeyNotFound((bucket, key))
  
  return True


def GetBucketKeys(bucket, strict=False):
  """Returns the keys in a bucket.
  
  Args:
    bucket: string, name of bucket for state variables
  
  NOTE(g): Having a "GetKeys()" function is rejected, due to it being confusing.
  """
  global SHARED_STATE
  
  if not BucketExists(bucket):
    # If we are strict, raise an exception...
    if strict:
      raise StateBucketDoesntExist('Bucket doesnt exist: %s' % bucket)
    
    # Else, return empty list, so we keep rollin
    else:
      return []
  
  # ThreadSafeDict protects
  keys = SHARED_STATE[bucket].keys()
  
  return keys


def GetBucketData(bucket, default=None):
  """Returns the data in a bucket.
  
  Args:
    bucket: string, name of bucket for state variables
  
  NOTE(g): Having a "GetKeys()" function is rejected, due to it being confusing.
  """
  global SHARED_STATE
  
  if not BucketExists(bucket):
    if default == None:
      raise StateBucketDoesntExist('Bucket doesnt exist: %s' % bucket)
    else:
      return default
  
  # Return the data
  return SHARED_STATE[bucket]


def Load(bucket, path, clear=True, critical=False):
  """TODO(g): Figure out how the names work.  I like this format as well.
      I guess I could just leave them both, and people could pick what they
      want to use, but I dont like having repitition...
  """
  return ImportSave(bucket, path, clear=clear, critical=critical)


def ImportSave(bucket, save, clear=True, critical=False, imported_data=None):
  """Imports the saved data file.  Tests for presence of "%(key)s" to load
  a single file, or process all the files.
  
  If a single file is to be loaded, and it does not contain keys (not a dict
  formated YAML file), then all the data will be stored in the key "default".
  
  Args:
    bucket: string, bucket to import into
    save: string, path to load save YAML file from
    clear: boolean (default=True), if True, this will clear the current bucket
        before loading the new data
    critical: boolean (default=False), if True, and there is an error loading
        the save file, then an Exception will be raised.  If False, then the
        function returns and execution carries on without the state loaded.
    imported_data: any (default=None), if set, this will become the new data.
        Why have this?  Because if you want to mark-up the data in your own
        way and yet use the rest of this function's features, this is why.
        Specially, procblock.procyaml.Import() does special cool things, but
        I want unidist to stand alone from procblock, so I am not going to
        import that function.
  
  TODO(g): Does this need a better name?  Im using it initially to populate the
      YAML files I saved set, back in...  Load?
  """
  # If we want to clear this first, then clear it
  if clear:
    _EnsureBucketExists(bucket)
    SHARED_STATE[bucket] = threadsafedict.ThreadSafeDict()
  
  # %(key)s == %s.  Make it so.  This way we can use both, which is convenient.
  if '%(keys)' in save:
    save = save.replace('%(key)s', '%s')
  
  # Load the data from the save file, a single save file, because no %s
  if '%s' not in save:
    if os.path.isfile(save):
      try:
        # If we dont have imported data passed to us, load it
        if not imported_data:
          fp = open(save, 'r')
          data = yaml.load(fp)
          fp.close()
        # Else, just use the imported data.  (See docstring for why we do this)
        else:
          data = imported_data
      except Exception, e:
        log('Failed to import file: %s: %s' % (save, e), logging.ERROR)
        
        # If this is a critical import, fail.  Otherwise just move on.
        if critical:
          raise Exception('Failed to import file: %s: %s' % (save, e))
        else:
          log('Failed to import file: %s: %s' % (save_path, e), logging.ERROR)
          return
      
      # If the data is in dict form, import it
      if type(data) == dict:
        #TODO(g): Do this in the second least naive way possible
        for key in data:
          Set(bucket, key, data[key])
          #log('ImportSave: %s: %s: %s: (key count=%s)' % (bucket, key, data[key], len(GetBucketKeys(bucket))))
      
      # Else, it's not the format I expect, but I'm loading it into default to do
      #   the best thing for general development purposes.  Test your scripts
      #   and blocks!  You have more than enough rope to hang yourself with this.
      else:
        #TODO(g): Remove this after testing.  This was a bad idea from the start
        #if 0:
        #  log('ImportSave data not dict type, saving to %s.default: %s' % (bucket, type(data)), logging.INFO)
        #  Set(bucket, 'default', data)
        
        # Ensure the bucket exists, no data yet
        _EnsureBucketExists(bucket)
    
    else:
      log('sharedstate import not found: %s' % save)
  
  # Else, load our state from any file that matches the substitution
  else:
    save_glob = save.replace('%s', '*')
    files = glob.glob(save_glob)
    
    save_regex = save.replace('%s', '(.*)')
    
    for filepath in files:
      found = re.findall(save_regex, filepath)
      if found:
        # The key is what we matched in the regex
        key = found[0]
        
        # This should be redundant, the glob just found it, but being safe
        if os.path.isfile(filepath):
          try:
            # If we dont have imported data passed to us, load it
            if not imported_data:
              fp = open(filepath, 'r')
              data = yaml.load(fp)
              fp.close()
            # Else, just use the imported data.  (See docstring for why we do this)
            else:
              data = imported_data
          except Exception, e:
            # If this is a critical import, fail.  Otherwise just move on.
            if critical:
              raise Exception('Failed to import file: %s: %s' % (save_path, e))
            else:
              log('Failed to import file: %s: %s' % (filepath, e), logging.ERROR)
              continue
          
          # Save all of this data into the key, we extracted with the save_regex
          log('Imported Bucket: %s  Key: %s:  %s' % (bucket, key, filepath))
          Set(bucket, key, data)


def Save(bucket, key=None, save=None, backups=2, delay=0, remote_node_sync=None,
         remote_node_url=None, remote_node_delay=30):
  """Save this bucket and key, as specified.
  
  Args:
    bucket: string, name of bucket for state variables
    key: string or any (optional), key to store data against.  Any valid dict key.
        If not specified, then all the keys will be saved.  This could be slow
        for key series data, so be aware.
    save: string (optional, default=None), the file path to save YAML files for
        this content.
        If "%(key)s" substring is present, each key in this bucket will be saved
        to it's own YAML file.  If it is not present, then all keys for a bucket
        will save into this one file.
    delay: number, seconds to delay in saving after a change, to incorporate
        many potential changes and minimize writes
    backups: int, the number of backups to keep of the saved file, in case of
        damage or other desire for change rollback.  Set backups=None to save,
        but not perform a backup
    remote_node_sync: list of strings (optional), list of hosts to remotely
        connect to on changes and replicate the changes
    remote_node_url: string, URL to RPC call for making this sync
    remote_node_delay: number, seconds to delay in replicating data after change
  """
  global SHARED_STATE
  
  import stack
  log(stack.Mini(5))
  
  # Ensure this bucket exists, we're going to create it
  _EnsureBucketExists(bucket)
  
  # If no save was specified, get the Registered save
  if save == None:
    save = GetRegisteredSave(bucket)
  
  # If no save file is specified, warn and return
  if not save:
    log('No save file specified: %s: %s' % (bucket, key), logging.WARN)
    return
  
  # %(key)s == %s.  Make it so.  This way we can use both, which is convenient.
  if '%s' in save:
    save = save.replace('%s', '%(key)s')
  
  #TODO(g): I want to save all the places we're saving these files, so they
  #   can be automatically loaded up again on starting the program, keeping
  #   all state, even if it not specified.  This is a good thing.  If we want
  #   to save it, we want to reload it on complete...
  pass
  
  # Get the last time it was saved
  if '%(key)s' in save:
    # If we have a key, we only save it
    if key:
      keys = [key]
    
    # Else, we dont have a key specified, so save all the keys
    else:
      keys = GetBucketKeys(bucket)
    
    # Process the one or more keys, from the list form
    for key in keys:
      shared_save_key = '%s.%s.' % (bucket.replace('.', '_'), key)
      last_saved = Get('__sharedstate.saved', shared_save_key, default=0)
      
      # If delay seconds has passed since it was last saved
      if not delay or time.time() - last_saved >= delay:
        save_final = save % {'key':run.MakePathSafeToCreate(key)}
        log('Saved sharedstate with key: %s: %s: %s' % (bucket, key, save_final))
        
        #TODO(g): Backups!  And Node Replication!
        temp_save = '%s.tmp' % save_final
        fp = open(temp_save, 'w')
        yaml.dump(SHARED_STATE[bucket][key], fp)
        fp.close()
        # Use the temporary file to avoid a terminate-while-saving error, where
        #   the program quits before the save is finished (threads!).  Safer.
        os.rename(temp_save, save_final)
  
  # Else, this is a single file storage, so just save it all
  else:
    shared_save_key = '%s.' % bucket.replace('.', '_')
    last_saved = Get('__sharedstate.saved', shared_save_key, default=0)
    
    #TODO(g): Backups!  And Node Replication!
    temp_save = '%s.tmp' % save
    fp = open(temp_save, 'w')
    yaml.dump(dict(SHARED_STATE[bucket]), fp)
    fp.close()
    
    # Use the temporary file to avoid a terminate-while-saving error, where
    #   the program quits before the save is finished (threads!).  Safer.
    os.rename(temp_save, save)
    
    log('Saved sharedstate: %s: %s' % (bucket, save))
  
  # Save when this data was updated
  Set('__sharedstate.saved', shared_save_key, time.time())


def Set(bucket, key, value, save=None, backups=2, delay=0,
        remote_node_sync=None, remote_node_url=None, remote_node_delay=30):
  """Lock the shared control access to locks.  This way we can safely acquire
  an individual lock or create a new lock without a race condition.
  
  Args:
    bucket: string, name of bucket for state variables
    key: string or any, key to store data against.  Any valid dict key.
    value: any, any object to be stored in shared state
    save: string (optional, default=None), the file path to save YAML files for
        this content.
        If "%(key)s" substring is present, each key in this bucket will be saved
        to it's own YAML file.  If it is not present, then all keys for a bucket
        will save into this one file.
    delay: number, seconds to delay in saving after a change, to incorporate
        many potential changes and minimize writes
    backups: int, the number of backups to keep of the saved file, in case of
        damage or other desire for change rollback.  Set backups=None to save,
        but not perform a backup
    remote_node_sync: list of strings (optional), list of hosts to remotely
        connect to on changes and replicate the changes
    remote_node_url: string, URL to RPC call for making this sync
    remote_node_delay: number, seconds to delay in replicating data after change
  """
  global SHARED_STATE
  
  # Ensure this bucket exists, we're going to create it
  _EnsureBucketExists(bucket)
  
  # If no save was specified, get the Registered save
  if save == None:
    save = GetRegisteredSave(bucket)
  
  # Set the bucket key to the value.  Safe via ThreadSafeDict
  SHARED_STATE[bucket][key] = value
  
  # If we want to save this result
  #TODO(g): Turn this into a common function, so it's not stuck in here...!!!
  if save:
    Save(bucket, key, save=save, backups=backups, delay=delay,
         remote_node_sync=remote_node_sync, remote_node_url=remote_node_url,
         remote_node_delay=remote_node_delay)
  
  # All functions should behave similarly, so this returns the value we set
  return value


def SetIfDoesntExist(bucket, key, value):
  """Returns boolean, whether the value was set or not.
  
  Args:
    bucket: string, name of bucket for state variables
    key: string or any, key to store data against.  Any valid dict key.
    value: any, any object to be stored in shared state
  """
  #TODO(g): Make totally atomic
  try:
    if not KeyExists(bucket, key):
      Set(bucket, key, value)
      return True
    else:
      return False
  except StateBucketDoesntExist, e:
    Set(bucket, key, value)
    return True


def Get(bucket, key, default=StateKeyNotFound):
  """Lock the shared control access to locks.  This way we can safely acquire
  an individual lock or create a new lock without a race condition.
  
  Args:
    bucket: string, name of bucket for state variables
    key: string or any, key to store data against.  Any valid dict key.
    default: any value (optional), if set and there is no data in this key,
        then this value will be set in the key and returned.
  
  NOTE(g): Do not try to implement the "strict=False" system for raising the
      Exception, it doesnt fit this function well, but using a default does
  """
  global SHARED_STATE
  
  if not BucketExists(bucket):
    # If we have a save file, but it hasnt been loaded yet, load it
    save_path = GetRegisteredSave(bucket)
    if save_path:
      ImportSave(bucket, save)
      
      # If the key now exists since we have loaded the state, return it
      if KeyExists(bucket, key):
        return SHARED_STATE[bucket][key]
    
    # If the default wasnt set, then raise an exception
    if default == StateKeyNotFound:
      raise StateBucketDoesntExist('Bucket doesnt exist: %s' % bucket)
    else:
      Set(bucket, key, default)
      return default
  
  # Else, If the key doesnt exist, but we have a default, then set the default
  elif not KeyExists(bucket, key):
    if default != StateKeyNotFound:
      Set(bucket, key, default)
      return default
  
  # Else, return the value
  else:
    return SHARED_STATE[bucket][key]


def GetSet(bucket, key, value):
  """Returns boolean if this named lock is locked.  If doesnt exist, still False
  
  Args:
    bucket: string, name of bucket for state variables
    key: string or any, key to store data against.  Any valid dict key.
    value: any, any object to be stored in shared state
  """
  global SHARED_STATE
  
  # Ensure this bucket exists, we're going to create it
  _EnsureBucketExists(bucket)
  
  # Set the bucket key to the value, and get original.  Safe via ThreadSafeDict
  previous_value = SHARED_STATE[bucket].GetSet(key, value)
  
  # Return the previous value, before we set this new value
  return previous_value

  
  
  
if __name__ == '__main__':
  #TODO(g): Turn into unittests
  print BucketExists('sofa')
  
  print Set('sofa', 'barn', 15)
  
  print BucketExists('sofa')
  
  print Get('sofa', 'barn')
  
  print Set('sofa', 'bush', 'Tundra!')
  
  print GetSet('sofa', 'bush', 'Tornado!')
  print GetSet('sofa', 'bush', 'Ohio!')
  
  print GetBucketKeys('sofa')
  
  print KeyExists('sofa', 'bush')
  print KeyExists('sofa', 'bush2')
  
  try:
    print KeyExists('sortaaaa', 'bush2')
  except StateBucketDoesntExist, e:
    print e


