"""
timeseries

By Geoff Howland  <geoff AT ge01f DOT com>

A wrapper for dealing with time series data.  Creates storage, stores, fetches,
and graphs.

Currently wrapped time series backends:

- RRDTool

TODO(g): Create more than a thin wrapper around RRD.  This is fine for now, but
it should be abstracted enough that any backend can be used, and even
alternative graphing methods could be used with RRD.
"""

import yaml
import os
import logging

from unidist.log import log

import rrd


# RRD defaults are kept here.  Extend as needed, or supply a different path
TIME_SERIES_DEFAULTS = 'timeseries_rrd_defaults.yaml'


def Create(filename, interval, create_fields, defaults_path=TIME_SERIES_DEFAULTS):
  """Create a time series file.
  
  Args:
    filename: string
    interval: int, seconds between saving time series data points
    create_fields: dict, fields and information about them for creation
    defaults_path: string (optional), path to YAML file with time series
        backend (ex. RRD) data
  
  Returns: int, status of command run (depends on backend, for now RRDTool code)
  """
  
  path = os.path.join(os.path.dirname(__file__), defaults_path)
  fp = open(path)
  defaults = yaml.load(open(path))
  fp.close()
  
  return rrd.CreateRrd(filename, interval, create_fields, defaults)
  

def GetLastUpdateTime(filename):
  """Returns the last time the RRD file was stored to, in epoch time (int).
  
  Args:
    filename: string, path to timeseries (RRD) file
  """
  return rrd.GetRrdLastUpdateTime(filename)


def Store(filename, interval, occurred, fields):
  """Stores data from the fields in the RRD, from the occurred time.
  
  Args:
    filename: string, path to timeseries (RRD) file
    internal: int, number of seconds between storage.  When created, the time
        series had this set, and this is required to ensure updates are only
        saved at the appropriate times.  RRD requires this, otherwise it's just
        good practice to have things aligned, even though it loses precision
        it's all fuzzy anyway.  Get used to it.
    occurred: int, the epoch time this data was collected.  Will be adjusted
        forward (important!) in time to the next even seconds by the interval.
    fields: dict, contains the data points which will be stored in the
        timeseries (RRD) file.  The dict keys should match the timeseries keys.
  """
  
  return rrd.StoreInRrd(filename, interval, occurred, fields)


def Fetch(filename, start_time, end_time=None):
  """Returns data stored in timeseries (RRD) file from start_time to now,
  or end_time, if specified.
  
  Args:
    filename: string, path to timeseries (RRD) file
    start_time: int, epoch time to start fetching from
    end_time: int (NOT YET IMPLMENTED), epoch time to stop fetching
  
  TODO(g): Implment end_time and other fetch parameteres.
  """
  return rrd.FetchFromRrd(filename, start_time)


def Graph(filename, image_filename, create_fields, graph_fields, method, title,
          label_vertical, defaults_path=TIME_SERIES_DEFAULTS, duration=600):
  """Graph a timeseries file.
  
  Args:
    filename: string, path to timeseries (RRD) file
    image_filename: string, path to render to
    create_fields: dict, the same data passed to the Create() method
    graph_fields: list of strings, ordered field names to graph
    method: string, valid: STACK, ... (TODO(g): Implement more!)
    title: string, title of graph
    label_vertical: string, label written vertically on left of graph
    defaults_path: string (optional), override time series defaults
  """
  
  path = os.path.join(os.path.dirname(__file__), defaults_path)
  fp = open(path)
  defaults = yaml.load(fp)
  fp.close()
  
  return rrd.GraphRrd(filename, image_filename, create_fields, graph_fields, method, title, label_vertical, defaults, duration=duration)


def GetNodeDataPaths(path):
  """Returns all the Time Series data (RRD) paths.
  
  The expected directory structure is:  path/NODENAME/TSDATANAME
  
  Example:  logs/ts/monkey.ge01f.com/cpu.rrd
            path='log/ts'
            node name='monkey.ge01f.com'
            TS data name='cpu.rrd'
  
  returns: list of strings, the relative path names, prefixed by path input arg
  """
  data_paths = []
  
  try:
    dirs = os.listdir(path)
  except OSError, e:
    log('Directory does not exist, creating: %s' % path, logging.WARN)
    #TODO(g):SECURITY: Set permissions...
    os.mkdir(path)
    return []
  
  for dir in dirs:
    node_path = '%s/%s' % (path, dir)
    if os.path.isdir(node_path):
      files = os.listdir(node_path)
      for name in files:
        file_name = '%s/%s' % (node_path, name)
        #TODO(g): Will have to change when a recond timeseries backend is added
        if os.path.isfile(file_name) and name.endswith('.rrd'):
          data_paths.append(file_name)
  
  return data_paths


if __name__ == '__main__':
  import os, time

  #TODO(g): Turn into unittests

  filename = 'cpu_test.rrd'
  image_filename = 'cpu_test.png'
  
  interval = 5

  # Create the RRD
  create_fields = {}
  #create_fields['user'] = {'type':'GAUGE', 'range_top':'U', 'range_bottom':0, 'heartbeat':120}
  create_fields['user'] = {'type':'GAUGE', 'range_top':'U'}
  create_fields['system'] = {'type':'GAUGE', 'range_top':'U'}
  create_fields['idle'] = {'type':'GAUGE', 'range_top':'U'}
  
  # If the RRD doesnt exist, create it
  if not os.path.exists(filename):
    Create(filename, interval, create_fields)
  
  
  # Get last updated
  print 'Last Updated: %s' % GetLastUpdateTime(filename)

  # Insert data
  fields = {'user':68, 'system':22, 'idle':33}
  Store(filename, interval, time.time(), fields)
  
  fields = {'user':56, 'system':6, 'idle':38}
  Store(filename, interval, time.time()+5, fields)
  
  fields = {'user':36, 'system':16, 'idle':48}
  Store(filename, interval, time.time()+10, fields)

  # Get last updated
  print 'Last Updated: %s' % GetLastUpdateTime(filename)
  
  # Fetch the data back
  fields = ['system', 'user', 'idle']
  start_time = -60 # 60 seconds from now
  fetched = Fetch(filename, start_time)
  print 'Fetched:'
  import pprint
  pprint.pprint(fetched)
  
  # Graph the RRD
  method = 'STACK'
  title = 'CPU Test'
  label_vertical = '% Used'
  graph_fields = ['system', 'user', 'idle']
  Graph(filename, image_filename, create_fields, graph_fields, method, title, label_vertical)

