"""
Time Series: RRD

Implementation of timeseries backend methods for RRDTool.

TODO(g): Optionally ensure a ramdisk is created for the RRD file, so no disk
    performance is lost.  Create the file using less storage space, if you
    are capturing a lot of data points with fine granularity, or accept
    sacraficing the memory to remove any IO hits.
TODO(g): Add in adjustable information for saving space.  Set by MB or time,
    or either?  Return both, if so...
"""

import os
import time
import logging

from unidist.log import log
from unidist.run import Run


def GetRrdLastUpdateTime(filename):
  """Returns: int, epooch time, last time RRD was updated"""
  updated = 0
  
  if not os.path.isfile(filename):
    Exception('RRD Does not exist: %s' % filename)
  
  cmd = 'rrdtool last %s' % filename
  (status, output, output_error) = Run(cmd)
  
  if status == 0:
    updated = int(output.strip())
  else:
    log('RRD File failed to find last update: %s: %s' % (filename, output_error), logging.ERROR)
  
  return updated


#Data:  fields = {'user':42, 'system':11, 'idle':46}
def StoreInRrd(filename, interval, occurred, fields):
  if not os.path.isfile(filename):
    raise Exception('RRD file not found: %s' % filename)
  
  #log('Storing RRD Occurrance: %s: %s: %s' % (filename, occurred, fields))
  
  # Fix occurred time to sync with interval
  occurred_orig = occurred
  occurred = int(occurred)
  offset = occurred % interval
  if offset != 0:
    occurred -= offset
  
  # Start off the command
  cmd = 'rrdtool update %s %s:' % (filename, occurred)
  
  # Sort the columns
  columns = fields.keys()
  columns.sort()
  
  # Add each of the columns, in order (always)
  for column in columns:
    #
    if column in fields:
      value = fields[column]
      
      if value == None:
        cmd += 'U:'
      else:
        cmd += '%s:' % value
    
    else:
      cmd += 'U:'
  
  # Trim the trailing colon (:)
  cmd = cmd[:-1]
  
  #print cmd
  (status, output, output_error) = Run(cmd)
  #print 'cmd: %s: %s: %s' % (cmd, status, output)
  
  if output_error:
    log('Cmd: %s  Error: %s' % (cmd, output_error))
    
    # Get last update time
    rrd_last_update_time = GetRrdLastUpdateTime(filename)
    
    ## Update the instance with this data
    #data = {'last_collect_time':rrd_last_update_time + interval}
    
    #log('RRD MISS: Ignoring: A duplicate entry was attempted, but RRD does not allow it: Last: %s  Occured: %s  Fields: %s' % (rrd_last_update_time, occurred_orig, fields))


def CreateRrd(filename, interval, create_fields, defaults, start=0):
  log('Creating RRD file: %s: %s seconds: %s' % (filename, interval, create_fields.keys()))
  
  cmd = 'rrdtool create %s --start %s --step %s ' % (filename, int(start - interval), interval)

  # Create them in sorted order, so we can insert them properly
  columns = create_fields.keys()
  columns.sort()
  
  # Add DSs for each column (sorted so order is repeatable)
  for column in columns:
    column_data = create_fields[column]
    
    #DS:user:GAUGE:120:0:100 \
    # Update the default column data with the specified column data
    info = dict(defaults['column_default'])
    info.update(column_data)
    
    cmd += 'DS:%s:%s:%s:%s:%s ' % (column, info['type'], info['heartbeat'], info['range_bottom'], info['range_top'])
  
  # Add RRAs
  for rra in defaults['rra']:
    #RRA:MIN:0.5:10:1008 \
    #print rra
    #print len(rra)
    cmd += 'RRA:%s:%s:%s:%s ' % (rra[0], rra[1], rra[2], rra[3])
  
  #print cmd

  # Run the command
  (status, output, output_error) = Run(cmd)
  
  if output_error:
    log('Failed to created RRD: %s: %s' % (cmd, output_error))
  
  return status


def GraphRrd(filename, image_path, create_fields, graph_fields, method, title, label_vertical, defaults, duration=600):
  log('Graphing RRD: %s  (%s)' % (filename, duration))
  
  # Create FILENAME.part, so we can move it afterwards and avoid broken images
  temp_image_path = '%s.part' % image_path
  cmd = 'rrdtool graph %s ' % temp_image_path
  cmd += '--title "%s" ' % (title)
  #print rrd_graph_info
  cmd += '--vertical-label "%s" ' % label_vertical
  cmd += '-w 400 -h 100 '
  cmd += '--start -%ss ' % duration
  
  # Populate column data for use below
  #TODO(g): Replace stuff below with this version that works for all the cases
  columns = {}
  for column in graph_fields:
    column_data = dict(defaults['column_default'])
    column_data.update(create_fields[column])
    columns[column] = column_data
  
  range_bottom = columns.get('range_bottom', 0)
  if range_bottom != None:
    cmd += '-l %d ' % range_bottom
  
  graph_top = columns.get('graph_top', None)
  if graph_top != None:
    cmd += '-u %d ' % graph_top
  
  # DEF
  for column in graph_fields:
    column_data = dict(defaults['column_default'])
    column_data.update(create_fields[column])
    
    cmd += 'DEF:%s=%s:%s:AVERAGE ' % (column, filename, column)
  
  # CDEF
  line = 1
  line_stack = []
  for column in graph_fields:
    column_data = dict(defaults['column_default'])
    column_data.update(create_fields[column])
    #{'key': ['duration'], 'heartbeat': 120, 'range_top': 100, 'range_bottom': 0, 'type': 'GAUGE'}
    #print column_data
    
    # Add the column to the line stack
    line_stack.append(column)
    
    # Create the line items
    line_items = ','.join(line_stack)
    
    # Add the pluses so they stack
    line_items += ',+' * (len(line_stack)-1)
    
    cmd += 'CDEF:Ln%d=%s ' % (line, line_items)
    
    # Include the line
    line += 1
  
  # AREA
  line = 1
  line_stack = []
  for column in graph_fields:
    column_data = dict(defaults['column_default'])
    column_data.update(create_fields[column])
    #{'key': ['duration'], 'heartbeat': 120, 'range_top': 100, 'range_bottom': 0, 'type': 'GAUGE'}
    #print column_data
    
    # Add the column to the line stack
    line_stack.append(column)
    
    # Create the line items
    line_items = ','.join(line_stack)
    
    # Add the pluses so they stack
    line_items += ',+' * (len(line_stack)-1)
    
    # Label the lines
    if line == 1:
      line_label = column
    else:
      line_label = '%s:STACK' % column
    
    cmd += 'AREA:%s%s:%s ' % (column, defaults['colors']['area'][line-1], line_label)
    
    # Include the line
    line += 1
  
  # LINE
  line = 1
  line_stack = []
  for column in graph_fields:
    cmd += 'LINE1:Ln%d%s ' % (line, defaults['colors']['line'][line-1])
    
    # Include the line
    line += 1
  
  # Date COMMENT
  cmd += '"COMMENT:\\n" '
  cmd += '"COMMENT:%s\\n" ' % time.asctime(time.localtime()).replace(':', '\\:')
  
  # GRPINTs
  for column in graph_fields:
    cmd += '"GPRINT:%s:LAST:Last %s\\: %%2.1lf" ' % (column, column)
  
  
  #log(cmd)
  (status, _, _) = Run(cmd)
  if status == 0:
    # Move the temp over the original
    try:
      os.rename(temp_image_path, image_path)
    except Exception, e:
      log('Renaming the RRD graph from %s to %s failed: %s' % (temp_image_path, image_path, e))


def FetchFromRrd(filename, start_time, consolidation='AVERAGE'):
  """Returns dict of dicts, keyed on time, keyed on column names, then values."""
  cmd = 'rrdtool fetch %s %s -s %s' % (filename, consolidation, start_time)
  
  # Values returned
  data = {}
  
  # Run the command
  (status, output, output_error) = Run(cmd)
  
  if output_error:
    raise Exception('Failed to fetch from RRD: %s: %s' % (cmd, output_error))
  else:
    lines = output.split('\n')
    
    # Get column order
    header = lines[0].strip()
    while '  ' in header:
      header = header.replace('  ', ' ')
    columns = header.split(' ')
    #print 'Columns: %s' % columns
    
    for line in lines[1:]:
      if ':' in line:
        (occurred, values) = line.split(':', 1)
        values = values.strip().split(' ')
        
        # Prepare a dict item for the column values to be inserted into
        item = {}
        
        for count in range(0, len(columns)):
          number_value = float(str(values[count]))
          
          item[columns[count]] = number_value
        
        # Store these column data item in our return data
        data[occurred] = item
        
        #print '%s: %s' % (occurred, values)
        
        #raise Exception('TBD...')
  
  return data


def Test():
  import yaml
  fp = open('timeseries_rrd_defaults.yaml')
  defaults = yaml.load(fp)
  fp.close()

  filename = 'cpu_test.rrd'
  graph_filename = 'cpu_test.png'
  
  interval = 5

  # Create the RRD
  create_fields = {}
  create_fields['user'] = {'type':'GAUGE', 'range_top':'U', 'range_bottom':0, 'heartbeat':120}
  create_fields['system'] = {'type':'GAUGE', 'range_top':'U', 'range_bottom':0, 'heartbeat':120}
  create_fields['idle'] = {'type':'GAUGE', 'range_top':'U', 'range_bottom':0, 'heartbeat':120}
  
  # If the RRD doesnt exist, create it
  if not os.path.exists(filename):
    CreateRrd(filename, interval, create_fields, defaults)
  
  
  #Old_CreateRrd(rrd_data, group)

  # Get last updated
  print GetRrdLastUpdateTime(filename)

  
  # Insert data
  fields = {'user':68, 'system':22, 'idle':33}
  StoreInRrd(filename, interval, time.time(), fields)
  
  fields = {'user':56, 'system':6, 'idle':38}
  StoreInRrd(filename, interval, time.time()+5, fields)
  
  fields = {'user':36, 'system':16, 'idle':48}
  StoreInRrd(filename, interval, time.time()+10, fields)

  #Old_StoreInRrd(rrd_data, group, {'items':columns})

  # Get last updated
  print GetRrdLastUpdateTime(filename)
  
  # Fetch the data back
  if 1:
    fields = ['system', 'user', 'idle']
    start_time = -60 # 60 seconds from now
    fetched = FetchFromRrd(filename, fields, start_time)
    print 'Fetched'
    import pprint
    pprint.pprint(fetched)
    
    sys.exit(0)
  
  # Graph the RRD
  create_fields = {}
  create_fields['user'] = {'type':'GAUGE', 'range_top':'U', 'range_bottom':0, 'heartbeat':120}
  create_fields['system'] = {'type':'GAUGE', 'range_top':'U', 'range_bottom':0, 'heartbeat':120}
  create_fields['idle'] = {'type':'GAUGE', 'range_top':'U', 'range_bottom':0, 'heartbeat':120}
  method = 'STACK'
  title = 'CPU Test'
  label_vertical = '% Used'
  graph_fields = ['system', 'user', 'idle']
  GraphRrd(filename, graph_filename, create_fields, graph_fields, method, title, label_vertical, defaults)
  #GraphRrds()
  
  


if __name__ == '__main__':
  Test()
